@@ -33,7 +33,6 @@
 ///////////////////////////////////////////////////////////////////////////
 
 
-#include <tmpDir.h>
 
 #include <ImfTiledOutputFile.h>
 #include <ImfInputFile.h>
@@ -46,9 +45,12 @@
 #include <stdio.h>
 #include <assert.h>
 
+#include "tmpDir.h"
+
+using namespace OPENEXR_IMF_NAMESPACE;
 using namespace std;
-using namespace Imath;
-using namespace Imf;
+using namespace IMATH_NAMESPACE;
+
 
 namespace {
 
@@ -306,7 +308,7 @@ writeCopyReadMIP (const char fileName1[],
 		{
                     assert ((levels2[l])[y][x] == (levels1[l])[y][x]);
 
-		    if (comp != B44_COMPRESSION)
+		    if (comp != B44_COMPRESSION && comp != B44A_COMPRESSION)
 			assert ((levels2[l])[y][x] == (levels[l])[y][x]);
 		}
     }
@@ -458,7 +460,7 @@ writeCopyReadRIP (const char fileName1[],
                         assert ((levels2[ly][lx])[y][x] ==
                                 (levels1[ly][lx])[y][x]);
 
-			if (comp != B44_COMPRESSION)
+			if (comp != B44_COMPRESSION && comp != B44A_COMPRESSION)
 			    assert ((levels2[ly][lx])[y][x] ==
 				    (levels[ly][lx])[y][x]);
 
